#include "socket_vol.h"


typedef int (*msg_handler_t)(int, uint64_t);


msg_handler_t msg_handler_g[SOCKET_VOL_NUM_MSG_IDS];


size_t buf_size_g = 0;
unsigned char *buf_g = NULL;


static int
file_create_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    uint64_t tmp_len;
    unsigned flags;
    hid_t fcpl = -1;
    hid_t fapl = -1;
    hid_t file = -1;
    int64_t file_id;

    assert(msg_len > 12);

    /* Decode flags */
    p = buf_g;
    UINT32DECODE(p, flags);

    /* Decode fcpl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode fcpl */
    if((fcpl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* Decode fapl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len < msg_len);

    /* Decode fapl */
    if((fapl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* No need to copy the name, just pass the message buffer to H5Fcreate().
     * Just make sure the null terminator is present */
    assert(buf_g[msg_len - 1] == '\0');

    /* Create the file */
    if((file = H5Fcreate((char *)p, flags, fcpl, fapl)) < 0)
        ERROR("H5Fcreate failed");

    /* Calculate reply message length */
    msg_len = 8;

    /* Encode the file id */
    assert((size_t)msg_len < SOCKET_VOL_MSG_HEADER_LEN);
    file_id = (int64_t)file;
    p = buf_g;
    INT64ENCODE(p, file_id);

    /* Send the file id */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of file id failed");

    /* Close decoded ids */
    if(H5Pclose(fcpl) < 0)
        ERROR("H5Pclose of fcpl failed");
    if(H5Pclose(fapl) < 0)
        ERROR("H5Pclose of fapl failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 8;
    file_id = (int64_t)-1;
    p = buf_g;
    INT64ENCODE(p, file_id);
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    /* Close decoded ids */
    if(fcpl >= 0 && H5Pclose(fcpl) < 0)
        PRINT_ERROR("H5Pclose of fcpl failed");
    if(fapl >= 0 && H5Pclose(fapl) < 0)
        PRINT_ERROR("H5Pclose of fapl failed");

    return(FAIL);
} /* end file_create_handler() */


static int
file_open_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    uint64_t tmp_len;
    unsigned flags;
    hid_t fapl = -1;
    hid_t file = -1;
    int64_t file_id;

    assert(msg_len > 12);

    /* Decode flags */
    p = buf_g;
    UINT32DECODE(p, flags);

    /* Decode fapl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len < msg_len);

    /* Decode fapl */
    if((fapl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* No need to copy the name, just pass the message buffer to H5Fopen().
     * Just make sure the null terminator is present */
    assert(buf_g[msg_len - 1] == '\0');

    /* Open the file */
    if((file = H5Fopen((char *)p, flags, fapl)) < 0)
        ERROR("H5Fopen failed");

    /* Calculate reply message length */
    msg_len = 8;

    /* Encode the file id */
    assert((size_t)msg_len < SOCKET_VOL_MSG_HEADER_LEN);
    file_id = (int64_t)file;
    p = buf_g;
    INT64ENCODE(p, file_id);

    /* Send the file id */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of file id failed");

    /* Close decoded id */
    if(H5Pclose(fapl) < 0)
        ERROR("H5Pclose of fapl failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 8;
    file_id = (int64_t)-1;
    p = buf_g;
    INT64ENCODE(p, file_id);
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    /* Close decoded id */
    if(fapl >= 0 && H5Pclose(fapl) < 0)
        PRINT_ERROR("H5Pclose of fapl failed");

    return(FAIL);
} /* end file_open_handler() */


static int
file_close_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t file_id;

    assert(msg_len == 8);

    /* Decode file id */
    p = buf_g;
    INT64DECODE(p, file_id);

    /* CLose the file */
    if(H5Fclose((hid_t)file_id) < 0)
        ERROR("H5Fclose failed");

    /* Calculate reply message length */
    msg_len = 1;

    /* Encode the confirmation message */
    buf_g[0] = (unsigned char)0;

    /* Send the confirmation */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of confirmation failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 1;
    buf_g[0] = (unsigned char)-1;
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    return(FAIL);
} /* end file_close_handler() */


static int
group_create_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t obj_id;
    uint64_t tmp_len;
    hid_t lcpl = -1;
    hid_t gcpl = -1;
    hid_t gapl = -1;
    hid_t group = -1;
    int64_t group_id;

    assert(msg_len > 16);

    /* Decode object id */
    p = buf_g;
    INT64DECODE(p, obj_id);

    /* Decode lcpl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode lcpl */
    if((lcpl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* Decode gcpl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode gcpl */
    if((gcpl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* Decode gapl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len < msg_len);

    /* Decode gapl */
    if((gapl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* No need to copy the name, just pass the message buffer to H5Gcreate2().
     * Just make sure the null terminator is present */
    assert(buf_g[msg_len - 1] == '\0');

    /* Create the group */
    if((group = H5Gcreate2((hid_t)obj_id, (char *)p, lcpl, gcpl, gapl)) < 0)
        ERROR("H5Gcreate2 failed");

    /* Calculate reply message length */
    msg_len = 8;

    /* Encode the group id */
    assert((size_t)msg_len < SOCKET_VOL_MSG_HEADER_LEN);
    group_id = (int64_t)group;
    p = buf_g;
    INT64ENCODE(p, group_id);

    /* Send the group id */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of group id failed");

    /* Close decoded ids */
    if(H5Pclose(lcpl) < 0)
        ERROR("H5Pclose of lcpl failed");
    if(H5Pclose(gcpl) < 0)
        ERROR("H5Pclose of gcpl failed");
    if(H5Pclose(gapl) < 0)
        ERROR("H5Pclose of gapl failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 8;
    group_id = (int64_t)-1;
    p = buf_g;
    INT64ENCODE(p, group_id);
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    /* Close decoded ids */
    if(lcpl >= 0 && H5Pclose(lcpl) < 0)
        PRINT_ERROR("H5Pclose of lcpl failed");
    if(gcpl >= 0 && H5Pclose(gcpl) < 0)
        PRINT_ERROR("H5Pclose of gcpl failed");
    if(gapl >= 0 && H5Pclose(gapl) < 0)
        PRINT_ERROR("H5Pclose of gapl failed");

    return(FAIL);
} /* end group_create_handler() */


static int
group_open_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t obj_id;
    uint64_t tmp_len;
    hid_t gapl = -1;
    hid_t group = -1;
    int64_t group_id;

    assert(msg_len > 16);

    /* Decode object id */
    p = buf_g;
    INT64DECODE(p, obj_id);

    /* Decode gapl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len < msg_len);

    /* Decode gapl */
    if((gapl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* No need to copy the name, just pass the message buffer to H5Gcreate2().
     * Just make sure the null terminator is present */
    assert(buf_g[msg_len - 1] == '\0');

    /* Create the group */
    if((group = H5Gopen2((hid_t)obj_id, (char *)p, gapl)) < 0)
        ERROR("H5Gopen2 failed");

    /* Calculate reply message length */
    msg_len = 8;

    /* Encode the group id */
    assert((size_t)msg_len < SOCKET_VOL_MSG_HEADER_LEN);
    group_id = (int64_t)group;
    p = buf_g;
    INT64ENCODE(p, group_id);

    /* Send the group id */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of group id failed");

    /* Close decoded id */
    if(H5Pclose(gapl) < 0)
        ERROR("H5Pclose of gapl failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 8;
    group_id = (int64_t)-1;
    p = buf_g;
    INT64ENCODE(p, group_id);
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    /* Close decoded id */
    if(gapl >= 0 && H5Pclose(gapl) < 0)
        PRINT_ERROR("H5Pclose of gapl failed");

    return(FAIL);
} /* end group_open_handler() */


static int
group_close_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t group_id;

    assert(msg_len == 8);

    /* Decode group id */
    p = buf_g;
    INT64DECODE(p, group_id);

    /* CLose the group */
    if(H5Gclose((hid_t)group_id) < 0)
        ERROR("H5Gclose failed");

    /* Calculate reply message length */
    msg_len = 1;

    /* Encode the confirmation message */
    buf_g[0] = (unsigned char)0;

    /* Send the confirmation */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of confirmation failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 1;
    buf_g[0] = (unsigned char)-1;
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    return(FAIL);
} /* end group_close_handler() */


static int
dataset_create_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t obj_id;
    uint64_t tmp_len;
    hid_t dtype = -1;
    hid_t space = -1;
    hid_t lcpl = -1;
    hid_t dcpl = -1;
    hid_t dapl = -1;
    hid_t dset = -1;
    int64_t dset_id;

    assert(msg_len > 16);

    /* Decode object id */
    p = buf_g;
    INT64DECODE(p, obj_id);

    /* Decode datatype length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode datatype */
    if((dtype = H5Tdecode(p)) < 0)
        ERROR("H5Tdecode failed");
    p += tmp_len;

    /* Decode dataspace length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode dataspace */
    if((space = H5Sdecode(p)) < 0)
        ERROR("H5Sdecode failed");
    p += tmp_len;

    /* Decode lcpl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode lcpl */
    if((lcpl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* Decode dcpl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode dcpl */
    if((dcpl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* Decode dapl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len < msg_len);

    /* Decode dapl */
    if((dapl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* No need to copy the name, just pass the message buffer to H5Dcreate2().
     * Just make sure the null terminator is present */
    assert(buf_g[msg_len - 1] == '\0');

    /* Create the dataset */
    if((dset = H5Dcreate2((hid_t)obj_id, (char *)p, dtype, space, lcpl, dcpl, dapl)) < 0)
        ERROR("H5Dcreate2 failed");

    /* Calculate reply message length */
    msg_len = 8;

    /* Encode the dataset id */
    assert((size_t)msg_len < SOCKET_VOL_MSG_HEADER_LEN);
    dset_id = (int64_t)dset;
    p = buf_g;
    INT64ENCODE(p, dset_id);

    /* Send the dataset id */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of dataset id failed");

    /* Close decoded ids */
    if(H5Tclose(dtype) < 0)
        ERROR("H5Tclose failed");
    if(H5Sclose(space) < 0)
        ERROR("H5Sclose failed");
    if(H5Pclose(lcpl) < 0)
        ERROR("H5Pclose of lcpl failed");
    if(H5Pclose(dcpl) < 0)
        ERROR("H5Pclose of dcpl failed");
    if(H5Pclose(dapl) < 0)
        ERROR("H5Pclose of dapl failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 8;
    dset_id = (int64_t)-1;
    p = buf_g;
    INT64ENCODE(p, dset_id);
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    /* Close decoded ids */
    if(dtype >= 0 && H5Tclose(dtype) < 0)
        PRINT_ERROR("H5Tclose failed");
    if(space >= 0 && H5Sclose(space) < 0)
        PRINT_ERROR("H5Sclose failed");
    if(lcpl >= 0 && H5Pclose(lcpl) < 0)
        PRINT_ERROR("H5Pclose of lcpl failed");
    if(dcpl >= 0 && H5Pclose(dcpl) < 0)
        PRINT_ERROR("H5Pclose of dcpl failed");
    if(dapl >= 0 && H5Pclose(dapl) < 0)
        PRINT_ERROR("H5Pclose of dapl failed");

    return(FAIL);
} /* end dataset_create_handler() */


static int
dataset_open_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t obj_id;
    uint64_t tmp_len;
    hid_t dapl = -1;
    hid_t dset = -1;
    int64_t dset_id;

    assert(msg_len > 16);

    /* Decode object id */
    p = buf_g;
    INT64DECODE(p, obj_id);

    /* Decode dapl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len < msg_len);

    /* Decode dapl */
    if((dapl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* No need to copy the name, just pass the message buffer to H5Dcreate2().
     * Just make sure the null terminator is present */
    assert(buf_g[msg_len - 1] == '\0');

    /* Open the dataset */
    if((dset = H5Dopen2((hid_t)obj_id, (char *)p, dapl)) < 0)
        ERROR("H5Dopen2 failed");

    /* Calculate reply message length */
    msg_len = 8;

    /* Encode the dataset id */
    assert((size_t)msg_len < SOCKET_VOL_MSG_HEADER_LEN);
    dset_id = (int64_t)dset;
    p = buf_g;
    INT64ENCODE(p, dset_id);

    /* Send the dataset id */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of dataset id failed");

    /* Close decoded ids */
    if(H5Pclose(dapl) < 0)
        ERROR("H5Pclose of dapl failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 8;
    dset_id = (int64_t)-1;
    p = buf_g;
    INT64ENCODE(p, dset_id);
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    /* Close decoded ids */
    if(dapl >= 0 && H5Pclose(dapl) < 0)
        PRINT_ERROR("H5Pclose of dapl failed");

    return(FAIL);
} /* end dataset_open_handler() */


static int
dataset_read_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t dset_id;
    uint64_t tmp_len;
    hid_t mem_type = -1;
    hid_t mem_space = -1;
    hid_t file_space = -1;
    hid_t dset_space = -1;
    hid_t dxpl = -1;
    size_t mem_type_elem_size;
    hssize_t mem_extent_npoints;
    H5S_sel_type mem_select_type;

    assert(msg_len > 16);

    /* Decode dataset id */
    p = buf_g;
    INT64DECODE(p, dset_id);

    /* Decode memory datatype length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode memory datatype */
    if((mem_type = H5Tdecode(p)) < 0)
        ERROR("H5Tdecode failed");
    p += tmp_len;

    /* Decode memory dataspace length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode memory dataspace if present, otherwise set to H5S_ALL */
    if(tmp_len) {
        if((mem_space = H5Sdecode(p)) < 0)
            ERROR("H5Sdecode failed");
        p += tmp_len;
    } /* end if */
    else
        mem_space = H5S_ALL;

    /* Decode file dataspace length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 <= msg_len);

    /* Decode file dataspace if present, otherwise set to H5S_ALL */
    if(tmp_len) {
        if((file_space = H5Sdecode(p)) < 0)
            ERROR("H5Sdecode failed");
        p += tmp_len;
    } /* end if */
    else
        file_space = H5S_ALL;

    /* Decode dxpl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len <= msg_len);

    /* Decode dxpl */
    if((dxpl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* If both dataspaces are H5S_ALL, we must send the app buffer size to the
     * client */
    if(mem_space == H5S_ALL && file_space == H5S_ALL) {
        assert(p - buf_g == msg_len);

        /* Get dataspace from dataset */
        if((dset_space = H5Dget_space((hid_t)dset_id)) < 0)
            ERROR("H5Dget_space failed");
        mem_select_type = H5S_SEL_ALL;

        /* Calculate application buffer size */
        if(0 == (mem_type_elem_size = H5Tget_size(mem_type)))
            ERROR("H5Tget_size failed");
        if((mem_extent_npoints = H5Sget_simple_extent_npoints(dset_space)) < 0)
            ERROR("H5Sget_simple_extent_npoints failed");
        tmp_len = (uint64_t)mem_type_elem_size * (uint64_t)mem_extent_npoints;

        /* Send application buffer size.  Note we can only recycle p and buf_g
         * here because we know we will enter the mem_select_type == H5S_SEL_ALL
         * branch below. */
        p = buf_g;
        UINT64ENCODE(p, tmp_len);
        if(send(client_sock, buf_g, 8, 0) != 8)
            ERROR("send of app buffer size failed");
    } /* end if */
    else
        /* Check if the memory selection is the same as the extent.  If not, we know
         * the buffer has been read from the client. */
        /* Note: This is a temporary solution until H5Dtransfer is implemented */
        if((mem_select_type = H5Sget_select_type(
                mem_space == H5S_ALL ? file_space : mem_space)) < 0)
            ERROR("H5Sget_select_type failed");

    if(mem_select_type == H5S_SEL_ALL) {
        if(!(mem_space == H5S_ALL && file_space == H5S_ALL)) {
            /* Calculate reply message length */
            if(0 == (mem_type_elem_size = H5Tget_size(mem_type)))
                ERROR("H5Tget_size failed");
            if((mem_extent_npoints = H5Sget_simple_extent_npoints(
                    mem_space == H5S_ALL ? file_space : mem_space)) < 0)
                ERROR("H5Sget_simple_extent_npoints failed");
            msg_len = 1 + (uint64_t)mem_type_elem_size * (uint64_t)mem_extent_npoints;
        } /* end if */
        else
            msg_len = 1 + tmp_len;

        /* Resize the buffer if necessary */
        if((size_t)msg_len > buf_size_g) {
            buf_size_g = (size_t)msg_len;
            if(NULL == (buf_g = (unsigned char *)realloc(buf_g, buf_size_g)))
                ERROR("realloc of buffer failed");
        } /* end if */

        p = &buf_g[1];
    } /* end if */
    else {
        assert(!(mem_space == H5S_ALL && file_space == H5S_ALL));

        /* Calculate reply message length */
        msg_len -= (p - buf_g) - 1;

#ifndef NDEBUG
        /* Make sure the message length is correct */
        if(0 == (mem_type_elem_size = H5Tget_size(mem_type)))
            ERROR("H5Tget_size failed");
        if((mem_extent_npoints = H5Sget_simple_extent_npoints(
                mem_space == H5S_ALL ? file_space : mem_space)) < 0)
            ERROR("H5Sget_simple_extent_npoints failed");
        assert(msg_len == 1 + (uint64_t)mem_type_elem_size * (uint64_t)mem_extent_npoints);
#endif /* NDEBUG */
    } /* end else */

    /* No need to copy the app buffer, just pass the message buffer to
     * H5Dread() */
    /* Read from the dataset */
    if(H5Dread((hid_t)dset_id, mem_type, mem_space, file_space, dxpl, p) < 0)
        ERROR("H5Dread failed");

    /* Encode the error status */
    p[-1] = (unsigned char)0;

    /* Send the data */
    if(send(client_sock, &p[-1], (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of data failed");

    /* Close decoded ids */
    if(H5Tclose(mem_type) < 0)
        ERROR("H5Tclose failed");
    if(mem_space != H5S_ALL && H5Sclose(mem_space) < 0)
        ERROR("H5Sclose failed");
    if(file_space != H5S_ALL && H5Sclose(file_space) < 0)
        ERROR("H5Sclose failed");
    if(dset_space >= 0 && H5Sclose(dset_space) < 0)
        ERROR("H5Sclose failed");
    if(H5Pclose(dxpl) < 0)
        ERROR("H5Pclose of lcpl failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 1;
    buf_g[0] = (unsigned char)-1;
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    /* Close decoded ids */
    if(mem_type >= 0 && H5Tclose(mem_type) < 0)
        PRINT_ERROR("H5Tclose failed");
    if(mem_space >= 0 && mem_space != H5S_ALL && H5Sclose(mem_space) < 0)
        PRINT_ERROR("H5Sclose failed");
    if(file_space >= 0 && file_space != H5S_ALL && H5Sclose(file_space) < 0)
        PRINT_ERROR("H5Sclose failed");
    if(dset_space >= 0 && H5Sclose(dset_space) < 0)
        ERROR("H5Sclose failed");
    if(dxpl >= 0 && H5Pclose(dxpl) < 0)
        PRINT_ERROR("H5Pclose of lcpl failed");

    return(FAIL);
} /* end dataset_read_handler() */


static int
dataset_write_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t dset_id;
    uint64_t tmp_len;
    hid_t mem_type = -1;
    hid_t mem_space = -1;
    hid_t file_space = -1;
    hid_t dset_space = -1;
    hid_t dxpl = -1;
    size_t mem_type_elem_size;
    hssize_t mem_extent_npoints;

    assert(msg_len > 16);

    /* Decode dataset id */
    p = buf_g;
    INT64DECODE(p, dset_id);

    /* Decode memory datatype length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode memory datatype */
    if((mem_type = H5Tdecode(p)) < 0)
        ERROR("H5Tdecode failed");
    p += tmp_len;

    /* Decode memory dataspace length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 < msg_len);

    /* Decode memory dataspace if present, otherwise set to H5S_ALL */
    if(tmp_len) {
        if((mem_space = H5Sdecode(p)) < 0)
            ERROR("H5Sdecode failed");
        p += tmp_len;
    } /* end if */
    else
        mem_space = H5S_ALL;

    /* Decode file dataspace length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len + 8 <= msg_len);

    /* Decode file dataspace if present, otherwise set to H5S_ALL */
    if(tmp_len) {
        if((file_space = H5Sdecode(p)) < 0)
            ERROR("H5Sdecode failed");
        p += tmp_len;
    } /* end if */
    else
        file_space = H5S_ALL;

    /* Decode dxpl length */
    UINT64DECODE(p, tmp_len);
    assert((p - buf_g) + tmp_len <= msg_len);

    /* Decode dxpl */
    if((dxpl = H5Pdecode(p)) < 0)
        ERROR("H5Pdecode failed");
    p += tmp_len;

    /* If both dataspaces are H5S_ALL, we must send the app buffer size to the
     * client */
    if(mem_space == H5S_ALL && file_space == H5S_ALL) {
        assert(p - buf_g == msg_len);

        /* Get dataspace from dataset */
        if((dset_space = H5Dget_space((hid_t)dset_id)) < 0)
            ERROR("H5Dget_space failed");

        /* Calculate application buffer size */
        if(0 == (mem_type_elem_size = H5Tget_size(mem_type)))
            ERROR("H5Tget_size failed");
        if((mem_extent_npoints = H5Sget_simple_extent_npoints(dset_space)) < 0)
            ERROR("H5Sget_simple_extent_npoints failed");
        tmp_len = (uint64_t)mem_type_elem_size * (uint64_t)mem_extent_npoints;

        /* Send application buffer size */
        p = buf_g;
        UINT64ENCODE(p, tmp_len);
        if(send(client_sock, buf_g, 8, 0) != 8)
            ERROR("send of app buffer size failed");

        /* Resize the buffer if necessary */
        if((size_t)tmp_len > buf_size_g) {
            buf_size_g = (size_t)tmp_len;
            if(NULL == (buf_g = (unsigned char *)realloc(buf_g, buf_size_g)))
                ERROR("realloc of buffer failed");
        } /* end if */

        /* Receive the application buffer */
        if((size_t)tmp_len != recv(client_sock, buf_g, (size_t)tmp_len, MSG_WAITALL))
            ERROR("recv of app buffer failed");
        p = buf_g;
    } /* end if */
    else
        assert(p - buf_g < msg_len);

    /* No need to copy the app buffer, just pass the message buffer to
     * H5Dwrite() */

    /* Write to the dataset */
    if(H5Dwrite((hid_t)dset_id, mem_type, mem_space, file_space, dxpl, p) < 0)
        ERROR("H5Dwrite failed");

    /* Calculate reply message length */
    msg_len = 1;

    /* Encode the confirmation message */
    buf_g[0] = (unsigned char)0;

    /* Send the confirmation */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of confirmation failed");

    /* Close decoded ids */
    if(H5Tclose(mem_type) < 0)
        ERROR("H5Tclose failed");
    if(mem_space != H5S_ALL && H5Sclose(mem_space) < 0)
        ERROR("H5Sclose failed");
    if(file_space != H5S_ALL && H5Sclose(file_space) < 0)
        ERROR("H5Sclose failed");
    if(H5Pclose(dxpl) < 0)
        ERROR("H5Pclose of lcpl failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 1;
    buf_g[0] = (unsigned char)-1;
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    /* Close decoded ids */
    if(mem_type >= 0 && H5Tclose(mem_type) < 0)
        PRINT_ERROR("H5Tclose failed");
    if(mem_space >= 0 && mem_space != H5S_ALL && H5Sclose(mem_space) < 0)
        PRINT_ERROR("H5Sclose failed");
    if(file_space >= 0 && file_space != H5S_ALL && H5Sclose(file_space) < 0)
        PRINT_ERROR("H5Sclose failed");
    if(dxpl >= 0 && H5Pclose(dxpl) < 0)
        PRINT_ERROR("H5Pclose of lcpl failed");

    return(FAIL);
} /* end dataset_write_handler() */


static int
dataset_close_handler(int client_sock, uint64_t msg_len)
{
    unsigned char *p = NULL;
    int64_t dset_id;

    assert(msg_len == 8);

    /* Decode dataset id */
    p = buf_g;
    INT64DECODE(p, dset_id);

    /* CLose the dataset */
    if(H5Dclose((hid_t)dset_id) < 0)
        ERROR("H5Dclose failed");

    /* Calculate reply message length */
    msg_len = 1;

    /* Encode the confirmation message */
    buf_g[0] = (unsigned char)0;

    /* Send the confirmation */
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        ERROR("send of confirmation failed");

    return(SUCCEED);

error:
    /* Send the error */
    msg_len = 1;
    buf_g[0] = (unsigned char)-1;
    if(send(client_sock, buf_g, (size_t)msg_len, 0) != (ssize_t)msg_len)
        PRINT_ERROR("send of error failed");

    return(FAIL);
} /* end dataset_close_handler() */


int
main(int argc, char *argv[])
{
    int server_sock;
    int client_sock;
    struct sockaddr_in server_addr;
    struct sockaddr_in client_addr;
    uint16_t server_port;
    socklen_t addr_len = sizeof(struct sockaddr_in);
    ssize_t recv_size;
    unsigned char *p = NULL;
    uint64_t msg_len;
    uint16_t msg_id;

    msg_handler_g[SOCKET_VOL_FILE_CREATE_ID] = file_create_handler;
    msg_handler_g[SOCKET_VOL_FILE_OPEN_ID] = file_open_handler;
    msg_handler_g[SOCKET_VOL_FILE_CLOSE_ID] = file_close_handler;
    msg_handler_g[SOCKET_VOL_GROUP_CREATE_ID] = group_create_handler;
    msg_handler_g[SOCKET_VOL_GROUP_OPEN_ID] = group_open_handler;
    msg_handler_g[SOCKET_VOL_GROUP_CLOSE_ID] = group_close_handler;
    msg_handler_g[SOCKET_VOL_DSET_CREATE_ID] = dataset_create_handler;
    msg_handler_g[SOCKET_VOL_DSET_OPEN_ID] = dataset_open_handler;
    msg_handler_g[SOCKET_VOL_DSET_READ_ID] = dataset_read_handler;
    msg_handler_g[SOCKET_VOL_DSET_WRITE_ID] = dataset_write_handler;
    msg_handler_g[SOCKET_VOL_DSET_CLOSE_ID] = dataset_close_handler;

    if(NULL == (buf_g = (unsigned char *)malloc(SOCKET_VOL_MSG_HEADER_LEN)))
        ERROR("malloc of buffer failed");
    buf_size_g = SOCKET_VOL_MSG_HEADER_LEN;

    assert(argc == 2);

    server_port = (uint16_t)atoi(argv[1]);

    if((server_sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        ERROR("socket failed");

    memset(&server_addr, 0, sizeof(server_addr));
    server_addr.sin_family = AF_INET;
    server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    server_addr.sin_port = htons(server_port);

    if(bind(server_sock, (struct sockaddr *)&server_addr, sizeof(server_addr)) < 0)
        ERROR("bind failed - port in use?");

    if(listen(server_sock, SOMAXCONN) < 0)
        ERROR("listen failed");

    do {
        if((client_sock = accept(server_sock, (struct sockaddr *)&client_addr, &addr_len)) < 0)
            ERROR("accept failed");

        if((recv_size = recv(client_sock, buf_g, SOCKET_VOL_MSG_HEADER_LEN, MSG_WAITALL)) < 0)
            ERROR("recv of message header failed");

        while(recv_size > 0) {
            assert(recv_size == SOCKET_VOL_MSG_HEADER_LEN);

            /* Decode the message header */
            p = buf_g;
            UINT16DECODE(p, msg_id);
            UINT64DECODE(p, msg_len);

            /* Resize the buffer if necessary */
            if((size_t)msg_len > buf_size_g) {
                buf_size_g = (size_t)msg_len;
                if(NULL == (buf_g = (unsigned char *)realloc(buf_g, buf_size_g)))
                    ERROR("realloc of buffer failed");
            } /* end if */

            /* Receive the remainder of the message */
            if((ssize_t)msg_len != recv(client_sock, buf_g, (size_t)msg_len, MSG_WAITALL))
                ERROR("recv of message body failed");

            /* Invoke the message handler */
            assert(msg_id < SOCKET_VOL_NUM_MSG_IDS);
            if(msg_handler_g[msg_id](client_sock, msg_len) < 0)
                ERROR("message handler failed");

            if((recv_size = recv(client_sock, buf_g, SOCKET_VOL_MSG_HEADER_LEN, MSG_WAITALL)) < 0)
                ERROR("recv of message header failed");
        } /* end while */

        close(client_sock);
    } while(1);

    close(server_sock);
    free(buf_g);
    buf_g = NULL;
    buf_size_g = 0;

    return(0);

error:
    close(server_sock);
    if(buf_g)
        free(buf_g);
    buf_g = NULL;
    buf_size_g = 0;

    return(1);
} /* end main() */

