#include <assert.h>
#include <hdf5.h>
#include "socket_vol.h"

int main(int argc, char *argv[]) {
    hid_t file, group1, group2, fapl, dcpl, space, space2, dset;
    hsize_t dims[2] = {2, 3};
    hsize_t cdims[2] = {2, 2};
    hsize_t mdims[2] = {H5S_UNLIMITED, H5S_UNLIMITED};
    hsize_t start[2] = {0, 0};
    hsize_t count[2] = {2, 3};
    float wbuf[2][3] = {{.1, .2, .3}, {.4, .5, .6}};
    float rbuf[2][3] = {{0., 0., 0.}, {0., 0., 0.}};
    float rbuf2[3][3] = {{0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}};
    float ebuf[2][3] = {{0., 0., 0.}, {0., 0., 0.}};
    int i, j;
    
    char *server_addr;
    uint16_t server_port;

    assert(argc == 3);

    server_addr = argv[1];
    server_port = (uint16_t)atoi(argv[2]);

    if((fapl = H5Pcreate(H5P_FILE_ACCESS)) < 0)
        assert(0 && "H5Pcreate");
    if(hdf5_socket_vol_set(fapl, server_addr, server_port) < 0)
        assert(0 && "H5VL_socket_vol_set");

    if((file = H5Fcreate("peekaboo.h5", H5F_ACC_TRUNC, H5P_DEFAULT, fapl)) < 0)
        assert(0 && "H5Fcreate");
    if((group1 = H5Gcreate2(file, "I", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0)
        assert(0 && "H5Gcreate2");
    if((group2 = H5Gcreate2(file, "see", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0)
        assert(0 && "H5Gcreate2");
    if(H5Gclose(group1) < 0)
        assert(0 && "H5Gclose");
    if(H5Fclose(file) < 0)
        assert(0 && "H5Fclose");
    if(H5Gclose(group2) < 0)
        assert(0 && "H5Gclose");

    if((file = H5Fopen("peekaboo.h5", H5F_ACC_RDWR, fapl)) < 0)
        assert(0 && "H5Fopen");
    if((group1 = H5Gcreate2(file, "you", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0)
        assert(0 && "H5Gcreate2");
    if(H5Gclose(group1) < 0)
        assert(0 && "H5Gclose");
    if(H5Fclose(file) < 0)
        assert(0 && "H5Fclose");

    if((file = H5Fopen("peekaboo.h5", H5F_ACC_RDWR, fapl)) < 0)
        assert(0 && "H5Fopen");
    if((group1 = H5Gopen2(file, "you", H5P_DEFAULT)) < 0)
        assert(0 && "H5Gopen2");
    if((group2 = H5Gcreate2(group1, "too", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0)
        assert(0 && "H5Gcreate2");

    if((space = H5Screate_simple(2, dims, mdims)) < 0)
        assert(0 && "H5Screate_simple");
    if((dcpl = H5Pcreate(H5P_DATASET_CREATE)) < 0)
        assert(0 && "H5Pcreate");
    if(H5Pset_chunk(dcpl, 2, cdims) < 0)
        assert(0 && "H5Pset_chunk");
    if(H5Pset_deflate(dcpl, 6) < 0)
        assert(0 && "H5Pset_deflate");
    if((dset = H5Dcreate2(group2, "now", H5T_NATIVE_DOUBLE, space, H5P_DEFAULT, dcpl, H5P_DEFAULT)) < 0)
        assert(0 && "H5Dcreate2");

    if(H5Dclose(dset) < 0)
        assert(0 && "H5Dclose");
    if(H5Pclose(dcpl) < 0)
        assert(0 && "H5Pclose");
    if(H5Gclose(group1) < 0)
        assert(0 && "H5Gclose");

    if((dset = H5Dopen2(group2, "now", H5P_DEFAULT)) < 0)
        assert(0 && "H5Dopen2");
    if(H5Dwrite(dset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, wbuf[0]) < 0)
        assert(0 && "H5Dwrite");
    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            ebuf[i][j] = wbuf[i][j];
    if(H5Dread(dset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, rbuf[0]) < 0)
        assert(0 && "H5Dread");
    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            assert(rbuf[i][j] == ebuf[i][j]);

    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            wbuf[i][j] += 1.;
    dims[0] = 3;
    if((space2 = H5Screate_simple(2, dims, mdims)) < 0)
        assert(0 && "H5Screate_simple");
    if(H5Sselect_hyperslab(space2, H5S_SELECT_SET, start, NULL, count, NULL) < 0)
        assert(0 && "H5Sselect_hyperslab");
    if((dset = H5Dopen2(group2, "now", H5P_DEFAULT)) < 0)
        assert(0 && "H5Dopen2");
    if(H5Dwrite(dset, H5T_NATIVE_FLOAT, space, H5S_ALL, H5P_DEFAULT, wbuf[0]) < 0)
        assert(0 && "H5Dwrite");
    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            ebuf[i][j] = wbuf[i][j];
    if(H5Dread(dset, H5T_NATIVE_FLOAT, space2, H5S_ALL, H5P_DEFAULT, rbuf2[0]) < 0)
        assert(0 && "H5Dread");
    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            assert(rbuf2[i][j] == ebuf[i][j]);
    if(H5Dread(dset, H5T_NATIVE_FLOAT, space, H5S_ALL, H5P_DEFAULT, rbuf[0]) < 0)
        assert(0 && "H5Dread");

    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            wbuf[i][j] += 1.;
    start[1] = 1;
    count[1] = 2;
    if(H5Sselect_hyperslab(space, H5S_SELECT_SET, start, NULL, count, NULL) < 0)
        assert(0 && "H5Sselect_hyperslab");
    if(H5Dwrite(dset, H5T_NATIVE_FLOAT, H5S_ALL, space, H5P_DEFAULT, wbuf[0]) < 0)
        assert(0 && "H5Dwrite");
    for(i=0; i<2; i++)
        for(j=1; j<3; j++)
            ebuf[i][j] = wbuf[i][j];
    if(H5Dread(dset, H5T_NATIVE_FLOAT, H5S_ALL, space, H5P_DEFAULT, rbuf[0]) < 0)
        assert(0 && "H5Dread");
    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            assert(rbuf[i][j] == ebuf[i][j]);

    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            wbuf[i][j] += 1.;
    dims[0] = 2;
    if(H5Sset_extent_simple(space2, 2, dims, mdims) < 0)
        assert(0 && "H5Sset_extent_simple");
    start[1] = 0;
    if(H5Sselect_hyperslab(space2, H5S_SELECT_SET, start, NULL, count, NULL) < 0)
        assert(0 && "H5Sselect_hyperslab");
    if(H5Dwrite(dset, H5T_NATIVE_FLOAT, space2, space, H5P_DEFAULT, wbuf[0]) < 0)
        assert(0 && "H5Dwrite");
    ebuf[0][1] = ebuf[0][0];
    ebuf[1][1] = ebuf[1][0];
    ebuf[0][2] = wbuf[0][0];
    ebuf[1][2] = wbuf[1][0];
    if(H5Dread(dset, H5T_NATIVE_FLOAT, space, space2, H5P_DEFAULT, rbuf[0]) < 0)
        assert(0 && "H5Dread");
    for(i=0; i<2; i++)
        for(j=0; j<3; j++)
            assert(rbuf[i][j] == ebuf[i][j]);

    if(H5Dclose(dset) < 0)
        assert(0 && "H5Dclose");
    if(H5Sclose(space) < 0)
        assert(0 && "H5Sclose");
    if(H5Sclose(space2) < 0)
        assert(0 && "H5Sclose");
    if(H5Gclose(group2) < 0)
        assert(0 && "H5Gclose");
    if(H5Fclose(file) < 0)
        assert(0 && "H5Fclose");

    if(H5Pclose(fapl) < 0)
        assert(0 && "H5Pclose");

    return(0);
}

